"""The inputs class is used to manage the inputs."""

from dataclasses import dataclass
from string import ascii_letters, digits, punctuation
import pygame
from .controls import Controls
from ..settings import Settings
_ACCEPTED_LETTERS = ascii_letters + digits + punctuation + " "

class Inputs:
    """
    The inputs class is used to manage the inputs.
    check if the user clicked somewhere or if a key have been pressed by using this class.
    """

    def __init__(self, settings: Settings) -> None:
        self.controls = Controls(settings)
        self.clear_mouse_velocity()
        self.event_list: list[pygame.event.Event] = []
        self.mouse_x = 0
        self.mouse_y = 0

    def update(self):
        """Get the current events."""
        self.event_list = pygame.event.get()

    def get_characters(self, extra_characters: str = ''):
        """Return all the letter characters a-z, digits 0-9, whitespace and punctuation."""
        return [
            event.unicode for event in self.event_list
            if event.type == pygame.KEYDOWN and event.unicode and event.unicode in _ACCEPTED_LETTERS + extra_characters
        ]

    @property
    def quit(self):
        """Return True if the user quited the pygame window."""
        return any(event.type == pygame.QUIT for event in self.event_list)

    def get_clicks(self, frame_abs_x: int = 0, frame_abs_y: int = 0):
        """Return the clicks and the position."""

        return {event.button if event.type in [pygame.MOUSEBUTTONDOWN, pygame.MOUSEBUTTONUP] else 0 :
            Click(event.pos[0] - frame_abs_x, event.pos[1] - frame_abs_y, event.type == pygame.MOUSEBUTTONUP)
            for event in self.event_list
            if event.type in [pygame.MOUSEBUTTONDOWN, pygame.MOUSEMOTION, pygame.MOUSEBUTTONUP]
        }

    def clear_mouse_velocity(self):
        """Remove the history of mouse velocity."""
        self.mouse_x = None
        self.mouse_y = None

    def get_keydown(self, key: int):
        """Return true if the key is just pressed down."""
        for event in self.event_list:
            if event.type == pygame.KEYDOWN and event.key == key:
                return True
        return False

    def get_keyup(self, key: int):
        """Return true if the key is just unpressed."""
        for event in self.event_list:
            if event.type == pygame.KEYUP and event.key == key:
                return True
        return False

    def get_mouse_velocity(self):
        """Return the current mouse speed."""
        for event in self.event_list:
            if event.type == pygame.MOUSEMOTION:
                if self.mouse_x is not None and self.mouse_y is not None:
                    velocity = event.pos[0] - self.mouse_x, event.pos[1] - self.mouse_y
                    self.mouse_x = event.pos[0]
                    self.mouse_y = event.pos[1]
                    return velocity
                self.mouse_x = event.pos[0]
                self.mouse_y = event.pos[1]
                return 0,0
        self.mouse_x = None
        self.mouse_y = None
        return 0,0

    def get_actions(self):
        """Return a dict of str: bool specifying if the action is trigger or not."""
        types = [event.key for event in self.event_list if event.type == pygame.KEYDOWN]
        return {
            action : any(int(key) in types for key in keys)
            for action, keys in self.controls.get_reverse_mapping().items()}

    def get_arrows(self):
        """Return the events involving an arrow."""
        return [
            (event.type, event.key) for event in self.event_list
            if hasattr(event, 'key') and event.key in [pygame.K_DOWN, pygame.K_UP, pygame.K_LEFT, pygame.K_RIGHT]
        ]

@dataclass
class Click:
    """Represent a click with the mouse."""

    x: int # The position of the mouse on the click
    y: int
    up: bool # True if it is a button up, False if it is a button down



"""The file module store all the file classes."""
from .font import FontFile, default_font
from .data import DataFile
from .image import ImageFile
from .music import MusicFile
from .sound import SoundFile
from .file import get_file
from .gif import GIFFile
__all__ = ['FontFile', 'default_font', 'MusicFile', 'SoundFile', 'ImageFile', 'DataFile', 'get_file', 'GIFFile']



"""Represent a sound in the assets/sounds folder."""

import json
import pygame
from .file import File, get_file


class SoundFile(File):
    """A SoundFile represent a sound in the assets/sounds folder."""

    def __init__(self, path: str) -> None:
        super().__init__(path)
        self.full_path = get_file('sounds', path)

        with open(get_file('sounds', 'categories.json'), 'r', encoding='utf-8') as f:
            categories: dict = json.load(f)
            if path in categories:
                self.category = categories[path]
            else:
                self.category = "unavailable"

    def get(self):
        """
        Get the sound from its file.
        
        Returns:
        ----
        sound: the sound as a pygame object
        category: str, the category of the sound to be played with the proper volume.
        """
        return pygame.mixer.Sound(self.full_path), self.category



"""A phase is one step of the game."""
from abc import ABC, abstractmethod
import pygame
from .screen.frame import Frame
from .game import Game
from .base import BaseRunnable
from .server import Server

class BasePhase(ABC):
    """
    A Phase is a step in the game. Each game should have a few phases.
    Exemple of phases: menus, lobby, stages, ...
    Create a subclass of phase to do whatever you need by
    rewriting the start, the __init__, _update and/or end method.
    If the game is online, you will need twice as much frames. One half for the Server and the other half for the game.
    For the server, don't use any frame, but use only the inputs from the players by using self.game.server.get_last_reception()
    and send data based on them to the players via self.game.server.send() (or .send_all()).
    """

    def __init__(self, name, runnable: BaseRunnable) -> None:
        """
        Create the phase.

        Runnable: the game or server instance
        """
        ABC.__init__(self)
        self._name = name
        self.runnable = runnable
        self.runnable.set_phase(name, self)

    @property
    def database(self):
        """Alias for self.game.database or self.server.database"""
        return self.runnable.database

    @property
    def logger(self):
        """Alias for self.game.database or self.server.logger"""
        return self.runnable.logger

    @property
    def config(self):
        """Alias for self.game.config or self.server.config"""
        return self.runnable.config

    @property
    def debug(self):
        """Alias for self.game.debug or self.server.debug"""
        return self.runnable.debug

    @abstractmethod
    def start(self, **kwargs):
        """This method is called at the start of the phase and might need several arguments."""
        raise NotImplementedError()

    @abstractmethod
    def loop(self, loop_duration: int):
        """This method is called at every loop iteration."""
        raise NotImplementedError()

    @abstractmethod
    def next(self):
        """
        If the phase is over, return the name of the next phase, if the phase is not, return an empty string.
        If it is the end of the game, return 'NO_NEXT'
        """
        raise NotImplementedError()

    @abstractmethod
    def update(self, loop_duration: int):
        """
        Update the phase based on loop duration, inputs and network (via the game instance)
        This method is called at every loop iteration.
        """
        raise NotImplementedError()

    @abstractmethod
    def end(self):
        """Action to do when the phase is ended."""
        raise NotImplementedError()

class ServerPhase(BasePhase, ABC):
    """The ServerPhase is a game phase to be add to the server only."""

    def __init__(self, name, server: Server) -> None:
        ABC.__init__(self)
        BasePhase.__init__(self, name, server)

    @property
    def server(self) -> Server:
        """Alias for the server."""
        return self.runnable

    @property
    def network(self):
        """Alias for self.server.network"""
        return self.server.network

    def loop(self, loop_duration: int):
        """Update the phase every loop iteraton."""
        self.update(loop_duration)

class GamePhase(BasePhase, ABC):
    """The ServerPhase is a game phase to be add to the game only."""

    def __init__(self, name, game: Game) -> None:
        ABC.__init__(self)
        BasePhase.__init__(self, name, game)
        self.frames: list[Frame] = []

        self.absolute_x = 0
        self.absolute_y = 0
        self.current_hover_surface = None

    def add_child(self, frame: Frame):
        """Add a new frame to the phase."""
        self.frames.append(frame)

    @property
    def game(self) -> Game:
        """Alias for the game."""
        return self.runnable

    @property
    def settings(self):
        """Alias for self.game.settings"""
        return self.game.settings

    @property
    def soundbox(self):
        """Alias for self.game.soundbox"""
        return self.game.soundbox

    @property
    def jukebox(self):
        """Alias for self.game.jukebox"""
        return self.game.jukebox

    @property
    def inputs(self):
        """Alias for self.game.inputs"""
        return self.game.inputs

    @property
    def network(self):
        """Alias for self.game.network"""
        return self.game.client

    def loop(self, loop_duration: int):
        """Update the phase."""
        self.update(loop_duration)
        self.__update_focus()
        self.__update_hover()
        for frame in self.frames:
            frame.loop(loop_duration)

    def __update_focus(self):
        """Update the focus of all the frames."""
        clicks = self.game.inputs.get_clicks()
        if 1 in clicks and not clicks[1].up:
            x = clicks[1].x
            y = clicks[1].y
            for frame in self.frames:
                if frame.x < x < frame.x + frame.width and frame.y < y < frame.y + frame.height:
                    frame.update_focus(x, y)
                else:
                    frame.remove_focus()

        actions = self.game.inputs.get_actions()
        if "next widget focus" in actions and actions["next widget focus"]:
            for frame in self.frames:
                if frame.focused:
                    frame.next_object_focus()

    def __update_hover(self):
        """Update the cursor and the over hover surface based on whether we are above one element or not."""
        clicks = self.inputs.get_clicks()
        is_one_hovered = False
        if 0 in clicks:
            x = clicks[0].x
            y = clicks[0].y
            for frame in self.frames:
                if frame.x < x < frame.x + frame.width and frame.y < y < frame.y + frame.height:
                    is_frame_hovered, surf = frame.update_hover(x, y)
                    if is_frame_hovered:
                        is_one_hovered = True
                        self.current_hover_surface = surf
                        break

        if not is_one_hovered:
            cursor = self.config.default_cursor
            if hasattr(pygame, cursor):
                cursor = getattr(pygame, cursor)
            pygame.mouse.set_cursor(cursor)

    @property
    def visible_frames(self):
        """Return all the visible frames sorted by increasing layer."""
        return sorted(filter(lambda f: f.visible, self.frames), key= lambda w: w.layer)

    def get_surface(self, width, height):
        """Return the surface to be displayed by the phase."""
        bg = pygame.Surface((width, height), pygame.SRCALPHA)
        for frame in self.visible_frames:
            surf = frame.get_surface()
            bg.blit(surf, (frame.x, frame.y))
        return bg



"""The Texts class is used to manage the texts of the game by returning strings taking automatically into account the language."""

from .database import Database
from ..settings import Settings


class Texts:
    """
    The class Texts is used to manage the texts of the game, that might be provided in several languages.
    """

    def __init__(self, database: Database, settings: Settings) -> None:
        self._db = database
        self._settings = settings
        self._last_language = settings.language
        texts_list = self._db.get_texts(self._last_language)
        self._text_dict = {pos : txt for pos, txt in texts_list[0]}

    def get_positions(self):
        """Return all the positions (text keys)."""
        return list(self._text_dict.keys())

    def get(self, position):
        """Return a piece of text."""
        if self._settings.language != self._last_language:
            self._last_language = self._settings.language
            texts_list = self._db.get_texts(self._last_language)
            self._text_dict = {pos : txt for pos, txt in texts_list[0]}

        if position in self._text_dict:
            return self._text_dict[position]
        return position



"""The game module contains the game class which is used to represent every game."""
import pygame
from .database import Texts, Speeches
from .database.database import GAME
from .connexion import Client
from .inputs import Inputs
from .settings import Settings
from .screen.screen import Screen
from .sound import SoundBox, Jukebox
from .base import BaseRunnable

class Game(BaseRunnable):
    """
    The game is the instance created and runned by the player.
    It can be online (with a server) or offline.
    """

    def __init__(self, online: bool = True, debug: bool = False) -> None:
        BaseRunnable.__init__(self, debug, GAME)
        pygame.init()

        self.settings = Settings()
        self.soundbox = SoundBox(self.settings)
        self.jukebox = Jukebox(self.settings)

        self.inputs = Inputs(self.settings)
        self._screen = Screen(self.config, self.settings)

        self.texts = Texts(self.database, self.settings)
        self.speeches = Speeches(self.database, self.settings)

        if online:
            self.client = Client(self.config)
        else:
            self.client = None
        self.online = online

    def update(self) -> bool:
        """Update all the component of the game."""
        loop_duration = self.clock.tick(self.config.get("max_frame_rate"))
        self.logger.update(loop_duration)
        self.inputs.update()
        self._screen.display_phase(self.phases[self.current_phase])
        self._screen.update()
        self.jukebox.update()
        if self.online:
            self.client.update()
        is_game_over = self.update_phases(loop_duration)
        return self.inputs.quit or is_game_over or (self.online and self.client.is_server_killed())



"""Image files represent the file of an image."""
import pygame
from .file import File, get_file

class ImageFile(File):
    """Represent the file of an image stored in the assets/images folder."""

    def __init__(self, path: str) -> None:
        super().__init__(path)
        self.full_path = get_file(folder='images', file=path)

    def get(self, size: tuple[int, int] | None = None, rotation: float = 0) -> pygame.Surface:
        """
        Get the Surface of the image.
        
        Params:
        ----
        size: tuple(int, int), the expected size of the return image. If no argument is passed, use the size of the image.
        rotation: float, Rotate the image of this amount, default is 0.
        """
        image = pygame.image.load(self.full_path).convert_alpha()
        surface = pygame.transform.rotate(image, rotation)
        if size is None:
            return surface
        else:
            return pygame.transform.scale(surface, size)



"""the element module contains the Element object, which is a base for every object displayed on the game window."""
from abc import ABC, abstractmethod
from typing import Optional
import pygame
from .animated_surface import AnimatedSurface

class Element(ABC):
    """Element is the abstract class for everything object displayed on the game window: widgets, actors, decors, frames."""

    def __init__(
        self,
        master,
        surface: AnimatedSurface | pygame.Surface,
        x: int,
        y: int,
        layer: int = 0,
        hover_surface: Optional[pygame.Surface] = None,
        hover_cursor: Optional[pygame.Cursor] = None,
        can_be_disabled: bool = True,
        can_be_focused: bool = True
    ) -> None:
        """
        Create an Element.

        Params:
        ----
        master: Frame or Phase, the master of this object.
        surface: The surface.
        x, y: the coordinates in the master.
        width, height: the dimension of the object.
        layer: the layer of the object. The smaller the more on the background
        image_duration (ms): If a list is provided as background, the background of the frame is changed every image_duration.
        if image_duration is a list, it must have the same length as background.
        in this case, the i-th image of the background will be displayed image_duration[i] ms.
        image_introduction: int, if you provided a list for the surface, the background will not cycle back to 0
        but to this index. 
        hover_surface: Surface. If a surface is provided, it to be displayed at the mouse location when the
        frame is hovered by the mouse.
        hover_cursor: Cursor. If a cursor is provided, it is the cursor of the mouse when the mouse is over the frame.
        ca_be_disabled, can_be_focused: Some element could be disabled, like widgets. Some element could be focused.
        """
        self.x = x
        self.y = y
        self.layer = layer

        self.visible = True
        self.can_be_focused = can_be_focused
        self.focused = False
        self.can_be_disabled = can_be_disabled
        self.disabled = False

        if isinstance(surface, pygame.Surface):
            self.surface = AnimatedSurface([surface], 2, 0)
        else:
            self.surface = surface

        self.width, self.height = self.surface.width, self.surface.height
        ABC.__init__(self)
        self.master = master
        self.master.add_child(self)

        self.hover_cursor = hover_cursor
        self.hover_surface = hover_surface

    @property
    def absolute_x(self):
        """Return the absolute value of x in the game window."""
        return self.master.absolute_x + self.x

    @property
    def absolute_y(self):
        """Return the absolute value of y in the game window."""
        return self.master.absolute_y + self.y

    @property
    def game(self):
        """Return the game."""
        return self.master.game

    def update_hover(self, hover_x, hover_y): #pylint: disable=unused-argument
        """Update the hover cursor and surface. To be overriden by element needing it."""
        return False, None

    @abstractmethod
    def get_surface(self) -> pygame.Surface:
        """Return the surface to be blitted."""
        raise NotImplementedError()

    def loop(self, loop_duration: int):
        """Update the element every loop iteration."""
        self.surface.update_animation(loop_duration)
        self.update(loop_duration)

    @abstractmethod
    def update(self, loop_duration: int):
        """Update the element logic every loop iteration."""
        raise NotImplementedError()

    def move(self, new_x: int, new_y: int):
        """Move the object."""
        self.x = new_x
        self.y = new_y

    def set_layer(self, new_layer: int):
        """Set a new value for the layer"""
        self.layer = new_layer

    def send_to_the_back(self):
        """Send the object one step to the back."""
        self.layer -= 1

    def send_to_the_front(self):
        """Send the object one step to the front."""
        self.layer += 1

    def hide(self):
        """Hide the object."""
        self.visible = False

    def show(self):
        """Show the object."""
        self.visible = True

    def enable(self):
        """Enable the object if it can be disabled."""
        if self.can_be_disabled:
            self.disabled = False

    def disable(self):
        """disable the object if it can be disabled."""
        if self.can_be_disabled:
            self.disabled = True

    def focus(self):
        """focus the object if it can be focused."""
        if self.can_be_focused:
            self.focused = False

    def unfocus(self):
        """Unfocus the object if it can be focused."""
        if self.can_be_focused:
            self.focused = False

    @property
    def relative_coordinate(self):
        """Return the coordinate of the element in its frame."""
        return (self.x, self.y)

    @property
    def absolute_coordinate(self):
        """Return the coordinate of the element in the game window"""
        return (self.absolute_x, self.absolute_y)

    @property
    def relative_rect(self):
        """Return the rect of the element in its frame."""
        return pygame.rect.Rect(self.x, self.y, self.width, self.height)

    @property
    def absolute_rect(self):
        """Return the rect of the element in the game window."""
        return pygame.rect.Rect(self.absolute_x, self.absolute_y, self.width, self.height)

    @property
    def shape(self):
        """Return the shape of the element"""
        return (self.width, self.height)

    @property
    def relative_right(self):
        """Return the right coordinate of the element in the frame."""
        return self.x + self.width

    @property
    def absolute_right(self):
        """Return the right coordinate of the element in the game window"""
        return self.absolute_x + self.width

    @property
    def relative_bottom(self):
        """Return the bottom coordinate of the element in the frame."""
        return self.y + self.height

    @property
    def absolute_bottom(self):
        """Return the bottom coordinate of the element in the game window."""
        return self.absolute_y + self.height

    @property
    def relative_left(self):
        """Return the left coordinate of the element in the frame."""
        return self.x

    @property
    def absolute_left(self):
        """Return the left coordinate of the element in the game window."""
        return self.absolute_x

    @property
    def relative_top(self):
        """Return the top coordinate of the element in the frame."""
        return self.y

    @property
    def absolute_top(self):
        """Return the top coordinate of the element in the game window."""
        return self.absolute_y



"""
The transition modules are used to make a transition from one phase to another
by extracting the arguments for the .start method of the new phase
"""
from typing import Any
from abc import ABC, abstractmethod
from .game import Game
from .server import Server
from .phase import GamePhase, ServerPhase

class GameTransition(ABC):
    """The GameTransition is to be used to make a transition between two GamePhases."""

    def __init__(self, game: Game, previous_phase: str, next_phase: str) -> None:
        ABC.__init__(self)
        game.set_transition(previous_phase, next_phase, self)

    @abstractmethod
    def apply(self, phase: GamePhase) -> dict[str, Any]:
        """
        Apply the transition between the two phases by extracting a dict from the ending phase.
        The keys of the dict are the argument name of the .start(...) method of the starting phase.
        The values are the values of the argument passed to the .start(...) method

        Example:
        >>> class Phase0(GamePhase):
        >>>     a: int
        >>>     b: float

        >>> class Phase1(GamePhase):            
        >>>     def start(self, a: int, b: float):
        >>>         ....

        >>> class Transition01(GameTransition):
        >>>    def apply(self, phase0: Phase0):
        >>>         retun {'a' : phase0.a, 'b': phase0.a + phase0.b**2}
        """
        raise NotImplementedError()

class ServerTransition(ABC):
    """The GameTransition is to be used to make a transition between two GamePhases."""

    def __init__(self, server: Server, previous_phase: str, next_phase: str) -> None:
        ABC.__init__(self)
        server.set_transition(previous_phase, next_phase, self)

    @abstractmethod
    def apply(self, phase: ServerPhase) -> dict[str, Any]:
        """
        Apply the transition between the two phases by extracting a dict from the ending phase.
        The keys of the dict are the argument name of the .start(...) method of the starting phase.
        The values are the values of the argument passed to the .start(...) method

        Example:
        >>> class Phase0(ServerPhase):
        >>>     a: int
        >>>     b: float

        >>> class Phase1(ServerPhase):            
        >>>     def start(self, a: int, b: float):
        >>>         ....

        >>> class Transition01(ServerTransition):
        >>>    def apply(self, phase0: Phase0):
        >>>         retun {'a' : phase0.a, 'b': phase0.a + phase0.b**2}
        """
        raise NotImplementedError()



"""
The database module contains the database class to interaxct with the database
the texts and speeches to display texts ad play sounds in the good language.
"""
from .database import Database
from .texts import Texts
from .speeches import Speeches
__all__ = ['Texts', 'Database', 'Speeches']



"""The Controls class is used to map inputs into mapper values."""

import pygame
from ..settings import Settings

class Controls:
    """
    The controls is used to map the keybord keys into actions.
    It contains a dictionnary of pygame.types : string, that map a type of event into a string.
    The string is an action, the pygame.type is a str(int).
    The current mapping is store in the dynamic data/keymap.json file.
    """

    def __init__(self, settings: Settings) -> None:

        self._key_map_dict: dict[str, str] = {}
        self._reverse_mapping = {}
        self._settings = settings
        self._previous_controls = self._settings.controls
        self._update_settings()

    def get_reverse_mapping(self):
        """
        Get the reverse control mapping.

        Returns:
        -----
        reverse_mapping: dict[str, list[str]]. The keys are the actions, the values the list of event that would trigger it.
        """
        if self._previous_controls != self._settings.controls:
            self._update_settings()
        return self._reverse_mapping

    def _update_settings(self):
        """Update the key map dict with the current settings."""
        controls = self._settings.controls
        self._key_map_dict = {}
        for key, action in controls.items():
            if not key.isdigit() and hasattr(pygame, key):
                self._key_map_dict[str(getattr(pygame, key))] = action
            else:
                self._key_map_dict[key] = action

        self._reverse_mapping = self._get_reversed_mapping()

    def _get_reversed_mapping(self):
        """Get all the defined keys and the actions."""
        reversed_mapping = {}
        for key, action in self._key_map_dict.items():
            if action in reversed_mapping:
                reversed_mapping[action].append(key)
            else:
                reversed_mapping[action] = [key]
        return reversed_mapping



"""This file contains the constants for the connexion module."""

# Ports and communication constants
DISCOVERY_PORT = 37020

# Communication protocol

HEADER = 'header'
NEW_ID = 'new_id'
ID = 'id'
EXIT = 'exit'
CONTENT = 'content'
BROADCAST_IP = 'broadcast_ip'
HOST_PORT = 'host_port'
TIMESTAMP = 'timestamp'
NEW_PHASE = 'new_phase'
DISCONNECTION = 'disconnection'

ONLINE = 'online'
OFFLINE = 'offline'



"""The Font module contain the font class."""
from pygame.font import Font as Ft
from pygame import Color
from pygame import Surface

class Font(Ft):
    """The Font class is used to display texts"""

    def __init__(
            self,
            name: str | None,
            size: int,
            color: Color,
            settings, bold: bool = False,
            italic: bool = False,
            underline: bool = False
        ) -> None:
        """
        Create a Font instance.

        Params:
        ----
        name: the path to the font in the assets/font folder.
        size: the size of the font
        color: the color of the font
        settings: the self.settings of the game
        bold, italic, underline: flags for the font.
        """
        super().__init__(name, size)
        self.name = name
        self.color = color
        self.set_bold(bold)
        self.set_italic(italic)
        self.set_underline(underline)
        self._settings = settings

    def render(self, text: str) -> Surface:
        """Create a surface from the font by writing a text."""
        return super().render(text, self._settings.antialias, self.color)

    def set_color(self, color: Color):
        """Change the color of the font."""
        self.color = color



"""Use this file to create a server."""
from typing import Any
from pygaming import Server, ServerPhase, ServerTransition

class ServerLobby(ServerPhase):
    
    def __init__(self, server: Server) -> None:
        super().__init__("lobby", server)
    
    def start(self):
        pass

    def _update(self, loop_duration: int):
        pass

    def next(self):
        return ''
    
    def end(self):
        self.network.send_all("match start", {"color1" : "blue", "color2" : "red"})

class ServerMacth(ServerPhase):

    def __init__(self, server: Server) -> None:
        super().__init__("match", server)

    def start(self, color1: str, color2: str):
        pass

    def _update(self, loop_duration: int):
        pass

    def next(self):
        return ''
    
    def end(self):
        pass

class ServerLobbyMatchTransition(ServerTransition):

    def __init__(self, server: Server) -> None:
        super().__init__(server, "lobby", "match")

    def apply(self, phase: ServerLobby) -> dict[str, Any]:
        return {"color1" : "blue", "color2" : "red"}

if __name__ == '__main__':
    server = Server(debug = False, nb_max_player = 6)
    ServerLobby(server)
    ServerMacth(server)
    ServerLobbyMatchTransition(server)
    server.run()


"""The Sound class is used to store sounds, the SoundBox class is used to manage them."""

from ..file import SoundFile
from ..settings import Settings

class Sound:
    """The Sound class represent a sound, it load the category and the file from a SoundFile object."""

    def __init__(self, sound_file: SoundFile) -> None:
        self._sound, self.category = sound_file.get()

    def set_volume(self, volume):
        """Set the volume of the osund"""
        self._sound.set_volume(volume)

    def play(self, loop, maxtime, fade_ms):
        """Play the sound once."""
        self._sound.play(loop, maxtime, fade_ms)

class SoundBox:
    """The Sound box is used to play all the sounds."""

    def __init__(self, settings: Settings) -> None:

        self._settings = settings

    def play_sound(self, sound: Sound, loop: int, maxtime: int, fade_ms: int):
        """Play the sound with the proper volume."""
        sound.set_volume(self._settings.volumes["sounds"][sound.category]*self._settings.volumes["main"])
        sound.play(loop, maxtime, fade_ms)



"""The server class is used to communicate with the clients."""

import socket
import threading
import json
import time
from ..config import Config
from ._constants import DISCOVERY_PORT,  CONTENT, HEADER, NEW_ID, ONLINE, OFFLINE, BROADCAST_IP, TIMESTAMP

class ClientSocketManager:
    """
    This class is used to store the client socked object along with its id, address and port.

    Without this class, deconnection of players might create duplicate ids.
    """

    def __init__(self, client_socket: socket.socket, id_: int, address: str, port: int):
        """Create an instance of the clientSocket."""
        self.socket = client_socket
        self.id_ = id_
        self.status = ONLINE
        self.address = address
        self.port = port

class Server:
    """
    The server must be unique. It is launched with the server_main function.
    Every player, i.e. every client, connect to this server. This server receive and
    transmit the data to the players.
    
    Params:
    -----
    nb_max_players: int, the maximum number of players in one game.
    """
    def __init__(self, config: Config, nb_max_player: int = 8):

        host_ip = socket.gethostbyname(socket.gethostname())
        self.config = config

        self._nb_max_player = nb_max_player
        self._server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._client_socket_managers: list[ClientSocketManager] = []
        self._running = True
        self._reception_buffer = []
        self.last_receptions = []
        print(f"Server launched: {host_ip}, {self.config.server_port}")
        self._server_socket.bind((host_ip, self.config.server_port))
        self._server_socket.listen(nb_max_player*2)
        threading.Thread(target=self._accept_clients).start()

        threading.Thread(target=self._broadcast_address, kwargs={'host_ip' : host_ip}).start()

    def _accept_clients(self):
        """Accept a new client."""
        while self._running:
            try:
                client_socket, (address, port) = self._server_socket.accept()
                if address not in [client_socket_m.address for client_socket_m in self._client_socket_managers]:
                    if self._client_socket_managers:
                        id_ = max(client_socket_m.id_ for client_socket_m in self._client_socket_managers) +1
                    else:
                        id_ = 1
                    self._client_socket_managers.append(ClientSocketManager(client_socket, id_, address, port))
                    print(f"New client connected: {address} has the id {id_}")
                else:
                    for client_socket_m in self._client_socket_managers:
                        if client_socket_m.address == address:
                            client_socket_m.status = ONLINE
                            client_socket_m.port = port
                            client_socket_m.socket = client_socket
                            print(f"Client {address} (id={id_}) is now reconnected")

                welcome_message = {HEADER : NEW_ID, CONTENT : id_}
                json_message = json.dumps(welcome_message)
                client_socket.send(json_message.encode())
                threading.Thread(target=self._handle_client, args=(client_socket, id_)).start()
            except OSError:
                print("Server disconnected.")
                self.stop()

    def _broadcast_address(self, host_ip: int):
        """Send in the socket.SOCK_DGRAM socket the host_ip and the host port every 5 seconds."""
        while self._running and self.get_nb_players() < self._nb_max_player:
            broadcast_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            broadcast_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            message = json.dumps({HEADER : BROADCAST_IP, CONTENT : host_ip})
            broadcast_socket.sendto(message.encode(), ('<broadcast>', DISCOVERY_PORT))
            time.sleep(5)  # Send broadcast every 5 seconds

    def _handle_client(self, client_socket: socket.socket, id_: int):
        while self._running:
            try:
                data = client_socket.recv(self.config.max_communication_length)
                if data:
                    json_data = json.loads(data.decode())
                    self._reception_buffer.append(json_data)
            except ConnectionError:
                for client_sck in self._client_socket_managers:
                    if client_sck.id_ == id_:
                        print(f"Client {client_sck.address} with id {client_sck.id_} just disconnected.")
                        client_sck.status = OFFLINE
                break

    def update(self) -> list[dict]:
        """Return the last data received."""
        self.last_receptions = self._reception_buffer.copy()
        self._reception_buffer.clear()

    def get_nb_players(self) -> int:
        """get the number of player connected to the server."""
        return len(list(filter(lambda csm: csm.status == ONLINE, self._client_socket_managers)))

    def is_player_online(self, id_) -> int:
        """Return True if the player with this id is currently online."""
        return len(list(filter(lambda csm: csm.status == ONLINE and csm.id_ == id_, self._client_socket_managers))) == 1

    def send(self, client_id, header, data):
        """The data to one client."""
        for client_socket in self._client_socket_managers:
            if client_socket.id_ == client_id and client_socket.status == ONLINE:
                json_data = json.dumps({HEADER : header, CONTENT : data, TIMESTAMP : time.time()})
                client_socket.socket.send(json_data.encode())

    def send_all(self, header, data):
        """Send data to all the clients."""
        for client_socket in self._client_socket_managers:
            if client_socket.status == ONLINE:
                try:
                    json_data = json.dumps({HEADER : header, CONTENT : data, TIMESTAMP : time.time()})
                    client_socket.socket.send(json_data.encode())
                except ConnectionResetError:
                    client_socket.status = OFFLINE

    def stop(self):
        """Stop the server when the process is finished."""
        self._running = False
        self._server_socket.close()
        for client_socket in self._client_socket_managers:
            client_socket.socket.close()

    def __del__(self):
        self.stop()



"""Use this file to create a game."""
from typing import Any
from pygaming import Game, GamePhase, GameTransition
from pygaming import HEADER, CONTENT

class GameLobby(GamePhase):
    
    def __init__(self, game: Game) -> None:
        super().__init__("lobby", game)
        self.is_ready = False
        self.chosen_color = "blue"
    
    def start(self):
        pass

    def _update(self, loop_duration: int):
        
        self.network.send("my choice", {"color" : self.chosen_color, "ready": self.is_ready})
        # Use inputs to choose a new color
        # Use screen to display choices
        # Use soundbox and Jukebox for the sounds and music
        # ....

    def next(self):
        return "match" if any(lr[HEADER] == "match start" for lr in self.network.last_receptions) else ''
    
    def end(self):
        pass

class GameMacth(GamePhase):

    def __init__(self, game: Game) -> None:
        super().__init__("match", game)

    def start(self, color1: str, color2: str):
        pass

    def _update(self, loop_duration: int):
        pass

    def next(self):
        return ''
    
    def end(self):
        pass

class GameLobbyMatchTransition(GameTransition):

    def __init__(self, game: Game) -> None:
        super().__init__(game, "lobby", "match")

    def apply(self, phase: GameLobby) -> dict[str, Any]:
        return {"color1" : phase.chosen_color, "color2" : "red"}

if __name__ == '__main__':
    game = Game(online = True, debug = False)
    GameLobby(game)
    GameMacth(game)
    GameLobbyMatchTransition(game)
    game.run()


"""
The install module is the code executed by the installer.
"""
import os
import json
import sys
import shutil

import tkinter as tk
from tkinter import filedialog

def ask_directory():
    """Ask for a directory."""
    root = tk.Tk()
    root.withdraw()  # Hide the main window
    folder_path = filedialog.askdirectory(initialdir=os.path.join(os.path.expanduser("~"), "Documents"))
    return folder_path

def install():
    """
    The function executed by the install-game.exe file to install the game.
    It already have the data, assets and src in the temporary folder sys._MEIPASS
    """
    # get the data
    #pylint: disable=protected-access
    base_path = sys._MEIPASS

    # Get the name
    config_path = os.path.join(base_path, 'data', 'config.json')

    with open(config_path, 'r', encoding='utf-8') as f:
        config = json.load(f)
        # Get the name of the game stored here by the build function
        name: str = config['name']

    # Ask for a place to place the game
    path_to_parent_folder = ask_directory()
    if path_to_parent_folder is None:
        return

    path_to_install = path_to_parent_folder + '/' + name
    # Create the folder where the data, assets and src will be store
    modified_path_to_install = path_to_install
    nb_copy = 1
    while os.path.exists(modified_path_to_install):
        modified_path_to_install = path_to_install + f' ({nb_copy})'
        nb_copy += 1

    print("The choosen folder is", modified_path_to_install)

    # save the config file with the folder where the data are saved.
    # This is useful then, bc you can move the .exe files without causing any issue.
    config['path'] = modified_path_to_install

    with open(config_path, 'w', encoding='utf-8') as f:
        json.dump(config, f)


    # Copy the data folder
    shutil.copytree(
        os.path.join(base_path, 'data'),
        modified_path_to_install + '/data'
    )
    print('The data folder has been copied.')
    # Copy the asset folder
    shutil.copytree(
        os.path.join(base_path, 'assets'),
        modified_path_to_install + '/assets'
    )
    print('The assets folder has been copied.')

    # Copy the src folder
    shutil.copytree(
        os.path.join(base_path, 'src'),
        modified_path_to_install + '/src'
    )
    print('The src folder has been copied.')

 # Copy the server
    if os.path.exists(os.path.join(base_path, 'server')):
        shutil.copytree(
            os.path.join(base_path, 'server'),
            os.path.join(modified_path_to_install, 'server'),
        )
        shutil.copyfile(
            os.path.join(modified_path_to_install, 'server', 'server.exe'),
            os.path.join(modified_path_to_install, f'{name}-server.exe'),
        )
        shutil.rmtree(os.path.join(modified_path_to_install, 'server'))
        print('The server.exe has been copied.')

    shutil.copytree(
        os.path.join(base_path, 'game'),
        os.path.join(modified_path_to_install, 'game'),
    )
    shutil.copyfile(
        os.path.join(modified_path_to_install, 'game', 'game.exe'),
        os.path.join(modified_path_to_install, f'{name}.exe'),
    )
    shutil.rmtree(os.path.join(modified_path_to_install, 'game'))
    print('The game.exe has been copied.')

if __name__ == '__main__':
    install()



"""
Pygaming is a python library used to create videogames in python, based on pygame.
Pygaming provide an exhaustive template to create your own offline and online games
by defining your own game phases and transitions between them. Pygaming also
provides settings, loggers, inputs, files, music and sounds, network, and database management,
as well as multi-language support, screen, widgets and frames.
"""
from .config import Config
from .font import Font
from .game import Game
from .base import NO_NEXT
from .logger import Logger
from .phase import ServerPhase, GamePhase
from .server import Server
from .settings import Settings
from .transition import ServerTransition, GameTransition

from .screen.screen import Screen
from .screen.frame import Frame
from .screen.element import Element
from .screen.animated_surface import AnimatedSurface, SurfaceLike, SurfacesLike

from .inputs import Inputs, Controls, Click
from .file import FontFile, DataFile, ImageFile, GIFFile, SoundFile, MusicFile, get_file
from .connexion import Client, Server as Network, HEADER, ID, CONTENT, TIMESTAMP

from .database import Database, Texts, Speeches
from . import commands
from .screen.colored_surface import ColoredSurface

__all__ = ['Config', 'Font', 'Game', 'NO_NEXT', 'Logger', 'ServerPhase', 'GamePhase',
           'Server', 'Settings', 'ServerTransition', 'GameTransition', 'Screen', 'Frame',
           'Element', 'AnimatedSurface', 'SurfaceLike', 'SurfacesLike', 'Inputs', 'Controls', 'Click', 'FontFile',
           'DataFile', 'ImageFile', 'GIFFile', 'SoundFile', 'MusicFile','get_file', 'Client',
           'Network', 'HEADER', 'ID', 'CONTENT', 'TIMESTAMP', 'Database', 'Texts', 'Speeches',
           'commands', 'ColoredSurface']



"""The server module contains the class Server."""
from .connexion import Server as Network, EXIT, NEW_PHASE
from .database.database import SERVER
from .base import BaseRunnable

class Server(BaseRunnable):
    """The Server is the instance to be run as a server for online game."""

    def __init__(self, debug: bool, nb_max_player: bool) -> None:
        """
        Create the Server.

        Params:
        debug: bool, if True, the database will not delete itself at the end and the logger will also log debug content.
        nb_max_player: The maximum number of player allowed to connect to the game.
        """
        super().__init__(debug, SERVER)
        self.network = Network(self.config, nb_max_player)

    def update(self):
        """Update the server."""
        loop_duration = self.clock.tick(self.config.get("max_frame_rate"))
        self.logger.update(loop_duration)
        self.network.update()
        previous = self.current_phase
        is_game_over = self.update_phases(loop_duration)
        if previous != self.current_phase:
            self.network.send_all(NEW_PHASE, self.current_phase)
        return is_game_over

    def stop(self):
        """Stop the event."""
        self.network.send_all(EXIT, '')
        self.network.stop()



"""
The Database is used to address queries to the database.
"""
import sqlite3 as sql
import os
from typing import Literal

from ..file.file import get_file
from ..config import Config

SERVER = 'server'
GAME = 'game'

class Database:
    """
    The Database instance is used to adress queries to the database.
    No need to have 2 databases on the same code as they will connect to the same db.
    The database automatically create a .sqlite file in the temporary folder /data/sql/db.sqlite,
    then execute every .sql file in the data/sql/ folder.
    At instance deletion, the .sqlite file is deleted if the debug mode is not selected ()
    
    """

    def __init__(self, config: Config, runnable_type: Literal['server', 'game'] = GAME, debug: bool=False) -> None:
        """
        Initialize an instance of Database.
        
        args:
        debug: when passed to true, the delation of the database is not done at the destruction of the instance.
        """
        self._debug = debug
        if runnable_type == SERVER:
            self._db_path = get_file('data','db-server.sqlite')
            self._table_path = get_file('data','sql-server/tables.sql')
            self._ig_queries_path = get_file('data', 'sql-server/ig_queries.sql', permanent=False)
            self._sql_folder = get_file('data', 'sql-server')
        if runnable_type == GAME:
            self._db_path = get_file('data', 'db-game.sqlite')
            self._table_path = get_file('data','sql-game/tables.sql')
            self._ig_queries_path = get_file('data', 'sql-game/ig_queries.sql', permanent=False)
            self._sql_folder = get_file('data', 'sql-game')

        # Remove the previous sqlite file if existing.
        if os.path.isfile(self._db_path):
            os.remove(self._db_path)

        # Create and connect to the sqlite file.
        self._conn = sql.connect(self._db_path)

        # Initialize the sqlite file with the tables.
        self._execute_sql_script(self._table_path)

        for root, _, files in os.walk(self._sql_folder):
            for file in files:
                complete_path = os.path.join(root, file).replace('\\', '/')
                if complete_path.endswith('.sql') and complete_path != self._table_path and complete_path != self._ig_queries_path:
                    if self._debug:
                        print(complete_path)
                    self._execute_sql_script(complete_path)

        # Execute the queries previously saved.
        self._execute_sql_script(self._ig_queries_path)
        self.default_language = config.default_language

    def _execute_select_query(self, query: str):
        """Execute a select query on the database."""
        if not query.startswith("SELECT"):
            print("The query is wrong:\n", query, "\nShould start by 'SELECT'")
        try:
            cur = self._conn.cursor()
            cur.execute(query)
            result = cur.fetchall()
            description = [descr[0] for descr in cur.description]
            cur.close()
            return result, description
        except sql.Error as error:
            print("An error occured while querying the database with:\n",query,"\n",error)
            return None, None

    def execute_insert_query(self, query: str):
        """Execute an insert query on the database."""
        if not query.startswith("INSERT INTO"):
            print("The query is wrong:\n", query, "\nShould start by 'INSERT INTO'")
        try:
            # Execute the query on the Database
            cur = self._conn.cursor()
            cur.execute(query)
            self._conn.commit()
            # Save the query on the ig_query file to execute it every time you launch the app.
            with open(self._ig_queries_path, 'a', encoding='utf-8') as f:
                f.write(";\n" + query)
            cur.close()
        except sql.Error as error:
            print("An error occured while querying the database with:\n",query,"\n",error)

    def _execute_sql_script(self, script_path: str):
        """Execute a script query on the database."""
        try:
            cur = self._conn.cursor()
            with open(script_path, 'r', encoding='utf-8') as f:
                script = f.read()
            if script:
                cur.executescript(script)
                self._conn.commit()
                cur.close()

        except sql.Error as error:
            print("An error occured while querying the database with the script located at\n",script_path,"\n",error)

    def __del__(self):
        """Destroy the Database object. Delete the database file"""
        self._conn.close()
        if os.path.isfile(self._db_path) and not self._debug:
            os.remove(self._db_path)

    def get_data_by_id(self, id_: int, table: str, return_id: bool = True):
        """Get all the data of one row based on the id and the table."""
        query = f"SELECT * FROM {table} WHERE {table}_id = {id_} LIMIT 1"
        result, description = self._execute_select_query(query)
        return {key : value for key,value in zip(description, result[0]) if (key != f"{table}_id" or return_id)}

    def get_collection_joined_by_id(self, id_, table: str, join_table: str, return_id: bool = True) -> list:
        """Get all the row of a the table 'table' having has foreign key for the table 'join_table' the id id_"""
        query = f"SELECT * FROM {table} WHERE {join_table}_id = {id_}"
        result, description = self._execute_select_query(query)
        return [{key : value for key,value in zip(description, res) if (key != f"{table}_id" or return_id)} for res in result]

    def get_texts(self, language: str):
        """Return all the texts of the game.
        If the text is not avaiable in the chosen language, get the text in the default language.
        """

        return self._execute_select_query(
            f"""SELECT position, text_value
                FROM localizations
                WHERE language_code = '{language}'

                UNION

                SELECT position, text_value
                FROM localizations
                WHERE language_code = '{self.default_language}'
                AND position NOT IN (
                    SELECT position
                    FROM localizations
                    WHERE language_code = '{language}'
            )"""
        )

    def get_speeches(self, language: str):
        """
        Return all the specches of the game of the given language.
        If the speech is not available in the given language, get it in the default language
        """

        return self._execute_select_query(
            f"""SELECT position, sound_path
                FROM speeches
                WHERE language_code = '{language}'

                UNION

                SELECT position, sound_path
                FROM speeches
                WHERE language_code = '{self.default_language}'
                AND position NOT IN (
                    SELECT position
                    FROM speeches
                    WHERE language_code = '{language}'
            )"""
        )



"""The build function create the game-install.exe file that is to be distributed."""
import os
import json
import shutil
import platform
import PyInstaller.__main__
from ..error import PygamingException

def build(name: str):
    """
    Build the game-install.exe file that need to be distributed, including a .zip of the src, assets and data files.
    Executing this file ask the user to choose a folder to save the game data, unzip them in this folder and
    then call pyinstaller to build the server and the game .exe files.
    This function must be called by using the command line `pygaming build [name-of-the-game]`.

    params:
    ---
    name: str, the name of the game.
    """
    # Create the sep for the add-data of pyinstaller
    if platform.system() == 'Windows':
        sep = ';'
    else:
        sep = ':'

    cwd = os.getcwd()
    config_path = os.path.join(cwd, 'data', 'config.json')
    this_dir = os.path.dirname(__name__)

    with open(config_path, 'r', encoding='utf-8') as f:
        config = json.load(f)
        config['name'] = name
    with open(config_path, 'w', encoding='utf-8') as f:
        json.dump(config, f)

    print("The config file has been modified successfully")

    cwd= os.getcwd()
    this_dir = os.path.dirname(__name__)

    game_options = [
        '--onefile',
        f"--icon={os.path.join(cwd, 'assets', 'icon.ico')}",
        f"--add-data={os.path.join(cwd, 'data')}{sep}data",
        f"--add-data={os.path.join(cwd, 'assets')}{sep}assets",
    ]

    installer_options = [
        '--onefile',
        #'--noconsole',
        f"--icon={os.path.join(cwd, 'assets', 'icon.ico')}",
        f"--add-data={os.path.join(cwd, 'data')}{sep}data",
        f"--add-data={os.path.join(cwd, 'assets')}{sep}assets",
        f"--add-data={os.path.join(cwd, 'src')}{sep}src",
    ]
    # Build the game file
    if os.path.exists(os.path.join(cwd, "src", "game.py")):
        command = [os.path.join(cwd, "src", "game.py")] + game_options + ['--windowed']
        PyInstaller.__main__.run(command)
        installer_options.append(f"--add-data={os.path.join(cwd, 'dist', 'game.exe')}{sep}game")
        print("The game has been built successfully")
    else:
        raise PygamingException("You need a game.py file as main file of the game")

    # Build the server file
    if os.path.exists(os.path.join(cwd, "src", "server.py")):
        command = [os.path.join(cwd, "src", "server.py")] + game_options
        PyInstaller.__main__.run(command)
        installer_options.append(f"--add-data={os.path.join(cwd, 'dist', 'server.exe')}{sep}server")
        print("The server has been built successfully")

    # Create the installer
    command = [os.path.join(os.path.abspath(this_dir), 'pygaming', 'commands/install.py')] + installer_options
    PyInstaller.__main__.run(command)

    # Copy paste it on the root.
    try:
        shutil.copyfile(
            os.path.join(cwd, 'dist/install.exe'),
            os.path.join(cwd, f'install-{name}.exe')
        )
    except FileNotFoundError:
        print("The file has been deleted by your antivirus, find it back and tell your antivirus it is ok")

    # Remove the .spec files
    if os.path.isfile(os.path.join(cwd, "game.spec")):
        os.remove(os.path.join(cwd, "game.spec"))
    if os.path.isfile(os.path.join(cwd, "server.spec")):
        os.remove(os.path.join(cwd, "server.spec"))
    if os.path.isfile(os.path.join(cwd, "install.spec")):
        os.remove(os.path.join(cwd, "install.spec"))



"""The client class is used to communicate with the server."""

import socket
import threading
import json
import time
from typing import Any
from ._constants import DISCOVERY_PORT, CONTENT, HEADER, ID, NEW_ID, BROADCAST_IP, TIMESTAMP, EXIT
from ..config import Config

class Client:
    """The Client instance is used to communicate with the server. It sends data via the .send()"""
    def __init__(self, config: Config):
        self._config = config
        self._reception_buffer = []
        self.last_receptions = []
        self.id = None
        server_ip = self._discover_server()
        self._connect_to_server(server_ip)

    def send(self, header: str, content: Any, ):
        """Send the content to the server, specifying the header."""
        message = {ID : self.id, HEADER : header, CONTENT : content, TIMESTAMP : time.time()}
        json_data = json.dumps(message)
        self.client_socket.send(json_data.encode())

    def _discover_server(self):
        """use the SOCK_DGRAM socket to discover the server ip"""
        discovery_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        discovery_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        discovery_socket.bind(('', DISCOVERY_PORT))
        while True:
            data, _addr = discovery_socket.recvfrom(self._config.max_communication_length)
            message = json.loads(data.decode())
            if HEADER not in message or message[HEADER] != BROADCAST_IP:
                continue
            server_ip = message[CONTENT]
            discovery_socket.close()
            return server_ip

    def _connect_to_server(self, server_ip):
        self.client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.client_socket.connect((server_ip, self._config.server_port))
        threading.Thread(target=self._receive).start()

    def _receive(self):
        while True:
            try:
                data = self.client_socket.recv(self._config.max_communication_length)
                if data:
                    json_data = json.loads(data.decode())
                    if json_data[HEADER] == NEW_ID:
                        self.id = json_data[CONTENT]
                    self._reception_buffer.append(json_data)
            except (ConnectionError, json.JSONDecodeError):
                self.close()
                break

    def close(self):
        """Close the client at the end of the process."""
        self._reception_buffer = [{HEADER : EXIT}]
        self.client_socket.close()

    def clean_last(self):
        """Clean the reception."""
        self._reception_buffer = {}

    def is_server_killed(self):
        """Verify if the server sent EXIT because it is killed."""
        return any(lr[HEADER] == EXIT for lr in self.last_receptions)

    def update(self):
        """Update the client every iteration with the last receptions."""
        self.last_receptions = self._reception_buffer.copy()
        self._reception_buffer.clear()

    def __del__(self):
        self.close()



"""The frame module contain the Frame class, base of all displayed object."""
import pygame
from .animated_surface import AnimatedSurface
from .element import Element

class Frame(Element):
    """
    The Frame represent a fraction of the screen.
    It has backgrounds and can contain many elements, including other frames, widgets and actors.
    """

    def __init__(
        self,
        master,
        x: int,
        y :int,
        background: AnimatedSurface | pygame.Surface,
        focus_background: AnimatedSurface | pygame.Surface = None,
        layer: int = 0,
        hover_surface: pygame.Surface | None = None,
        hover_cursor: pygame.Cursor | None = None,
    ) -> None:
        """
        Create the frame.

        Params:
        ----
        master: Another Frame or a phase.
        x, y: the coordinate of the top left of the frame, in its master.
        background: The AnimatedSurface or Surface representing the background of the Frame.
        focus_background: The AnimatedSurface or Surface representing the background of the Frame when it is focused.
        If None, copy the background
        layer: the layer of the frame on its master. Objects having the same master are blitted on it by increasing layer.
        hover_surface: Surface. If a surface is provided, it to be displayed at the mouse location when the
        frame is hovered by the mouse.
        hover_cursor: Cursor. If a cursor is provided, it is the cursor of the mouse when the mouse is over the frame.
        """
        self.children: list[Element] = []

        Element.__init__(
            self,
            master,
            background,
            x,
            y,
            layer,
            hover_surface,
            hover_cursor,
            can_be_disabled=False,
            can_be_focused=True
        )
        self.focused = False
        self._current_object_focus = None
        if focus_background is None:
            focus_background = background
        if isinstance(focus_background, pygame.Surface):
            self.focus_background = AnimatedSurface([focus_background], 4, 0)
        self.current_hover_surface = None

    def add_child(self, child: Element):
        """Add a new element to the child list."""
        self.children.append(child)

    def update_hover(self, hover_x, hover_y) -> tuple[bool, pygame.Surface | None]:
        """Update the hovering."""
        hover_x -= self.x
        hover_y -= self.y
        is_one_hovered = False
        for child in self.children:
            if child.visible:
                if child.x < hover_x < child.x + child.width and child.y < hover_y < child.y + child.height:
                    is_child_hovered, surf = child.update_hover(hover_x, hover_y)
                    if is_child_hovered:
                        is_one_hovered = True
                        self.current_hover_surface = surf
        return is_one_hovered, self.current_hover_surface

    def update_focus(self, click_x, click_y):
        """Update the focus of all the children in the frame."""
        click_x -= self.x
        click_y -= self.y
        self.focused = True
        self.surface.reset()
        one_is_clicked = False
        for (i,child) in enumerate(self.children):
            if child.visible and child.can_be_focused:
                if child.x < click_x < child.x + child.width and child.y < click_y < child.y + child.height:
                    child.focus()
                    self._current_object_focus = i
                    one_is_clicked = True
                else:
                    child.unfocus()
            else:
                child.unfocus()
        if not one_is_clicked:
            self._current_object_focus = None

    def next_object_focus(self):
        """Change the focused object."""
        if self._current_object_focus is None:
            self._current_object_focus = 0

        for element in self.children:
            if element.can_be_focused:
                element.unfocus()

        for i in range(1, len(self.children)):
            j = (i + self._current_object_focus)%len(self.children)
            if self.children[j].can_be_focused:
                self.children[j].focus()
                self._current_object_focus = j
                break

    def remove_focus(self):
        """Remove the focus of all the children."""
        self.focused = False
        self.focus_background.reset()
        for child in self.children:
            child.unfocus()

    def update(self, loop_duration: int):
        """Update all the children of the frame."""
        for element in self.children:
            element.loop(loop_duration)

    @property
    def visible_children(self):
        """Return the list of visible children sorted by increasing layer."""
        return sorted(filter(lambda ch: ch.visible, self.children), key= lambda w: w.layer)

    def get_surface(self):
        """Return the surface of the frame as a pygame.Surface"""
        if self.focused:
            background = self.focus_background.get()
        else:
            background = self.surface.get()
        for child in self.visible_children:
            x = child.relative_left
            y = child.relative_top
            surface = child.get_surface()
            background.blit(surface, (x,y))
        return background



"""The GifFile represent the file of a gif, then turned into a list of images and a gif speed."""

import pygame
from PIL import Image
from .file import get_file, File

class GIFFile(File):
    """A GifFile represent a GIF in the assets/images folder."""

    def __init__(self, path: str) -> None:
        super().__init__(path)
        self.full_path = get_file('images', path)

    def get(self, size: tuple[int, int] | None = None, rotation: float = 0):
        """Get the gif as a list of surfaces and frame durations."""
        gif = Image.open(self.full_path)
        gif.seek(0)
        images: list[pygame.Surface] = [pygame.image.fromstring(gif.convert('RGBA').tobytes(), gif.size, 'RGBA')]
        frame_durations = [gif.info['duration']]
        while True:
            try:
                gif.seek(gif.tell()+1)
                images.append(pygame.image.fromstring(gif.convert('RGBA').tobytes(), gif.size, 'RGBA'))
                frame_durations.append(gif.info['duration'])
            except EOFError:
                break

        images_to_return = []
        for im in images:
            surface = pygame.transform.rotate(im, rotation)
            if size is None:
                images_to_return.append(surface)
            else:
                images_to_return.append(pygame.transform.scale(im, size))
        return images_to_return, frame_durations



"""The config class is used to interact with the config file."""
import json
from .file import get_file

class Config:
    """The config class is used to interact with the config file."""

    def __init__(self) -> None:
        path = get_file('data', 'config.json', True)
        file = open(path, 'r', encoding='utf-8')
        self._data = json.load(file)
        file.close()

    def get(self, key: str):
        """Get the value of the config attribute"""
        if key in self._data:
            return self._data[key]
        return None

    @property
    def dimension(self):
        """Return the dimension of the window in px x px."""
        key = "screen"
        if key in self._data:
            return self._data[key]
        return (800, 600)

    @property
    def default_language(self):
        """Return the default language."""
        key = "default_language"
        if key in self._data:
            return self._data[key]
        return "en_US"

    @property
    def default_cursor(self):
        """Return the default cursor."""
        key = "default_cursor"
        if key in self._data:
            return self._data[key]
        return "SYSTEM_CURSOR_ARROW"

    @property
    def game_name(self):
        """Return the name of the game."""
        key = "name"
        if key in self._data:
            return self._data[key]
        return "MyGame"

    @property
    def server_port(self):
        """Return the server port of the game."""
        key = "server_port"
        if key in self._data:
            return self._data[key]
        return 50505

    @property
    def max_communication_length(self):
        """Return the maximum length of a communication of the game."""
        key = "max_communication_length"
        if key in self._data:
            return self._data[key]
        return 2048



"""The connexion module contains the client and server classes as well as some useful constants."""
from .client import Client
from .server import Server
from ._constants import ID, HEADER, CONTENT, EXIT, NEW_PHASE, DISCONNECTION, TIMESTAMP
__all__ = ['Client', 'Server', 'ID', 'HEADER', 'CONTENT', 'EXIT', 'NEW_PHASE', 'DISCONNECTION', 'TIMESTAMP']



"""A baseRunnable is an abstract object from which herit the game and the server."""
import sys
from abc import ABC, abstractmethod
from typing import Literal, Any
import pygame

from .logger import Logger
from .database import Database

from .config import Config
from .error import PygamingException

NO_NEXT = 'no_next'


class BaseRunnable(ABC):
    """The BaseRunnable Class is an abstract class for both the Game and the Server."""

    def __init__(self, debug: bool, runnable_type: Literal['server', 'game']) -> None:
        super().__init__()
        pygame.init()
        self.debug = debug
        self.config = Config()
        self.logger = Logger(self.config, debug)
        self.database = Database(self.config, runnable_type, debug)
        self.phases = {}
        self.transitions = {}
        self.current_phase = ""
        self.clock = pygame.time.Clock()

    @abstractmethod
    def update(self):
        """Update the runnable, must be overriden."""
        raise NotImplementedError()

    def set_phase(self, name: str, phase):
        """Add a new phase to the game."""
        if not self.phases:
            self.current_phase = name
        if name in self.phases:
            raise PygamingException("This name is already assigned to another frame.")
        self.phases[name] = phase
        return self

    def set_transition(self, phase1: str, phase2: str, transition):
        """
        Set a transition between two phases.

        Params:
        ----
        phase1: the name of the phase that is ending.
        phase2: the name of the phase that is starting.
        transition: the Transition object between the two phases.
        Returns:
        ----
        The game itself for method chaining
        """
        self.transitions[(phase1, phase2)] = transition
        return self

    def update_phases(self, loop_duration: int):
        """Update the phases of the game."""
        self.phases[self.current_phase].loop(loop_duration)
        next_phase = self.phases[self.current_phase].next()
        # Verify if the phase is over
        if next_phase and next_phase != NO_NEXT:
            # if it is, end the current phase
            self.phases[self.current_phase].end()
            # get the value for the arguments for the start of the next phase
            new_data = self.transitions[(self.current_phase, next_phase)].apply(
                self.phases[self.current_phase]
            )
            # change the phase
            self.current_phase = next_phase
            # start the new phase
            self.phases[self.current_phase].start(**new_data)

        # if NO_NEXT was return, end the game.
        return next_phase == NO_NEXT

    def stop(self):
        """Stop the algorithm properly."""

    def run(self, **kwargs0: dict[str, Any]):
        """Run the game."""
        stop = False
        self.phases[self.current_phase].start(**kwargs0)
        while not stop:
            stop = self.update()
        self.phases[self.current_phase].end()
        self.stop()
        pygame.quit()
        sys.exit(0)



"""The command line entry point for the pygaming commands."""
import sys
from .init_cwd import init_cwd
from .build import build

INIT_CWD = 'init'
BUILD = 'build'

def cli():
    """Execute a command called by the command line 'pygaming ...'"""  
    args = sys.argv
    if len(args) < 2:
        print(f'You need a command: {INIT_CWD} or {BUILD}')
        sys.exit(1)
    cmd = args[1]
    if cmd not in [INIT_CWD, BUILD]:
        print(f"invalid command, you need one of {INIT_CWD} or {BUILD} but got {cmd}")

    if cmd == INIT_CWD:
        init_cwd()
    if cmd == BUILD:
        if len(args) == 2:
            print("Please specify the name of your game as argument of 'pygaming build'")
        else:
            build(" ".join(word.capitalize() for word in args[2:]))



"""The colored_surface module contains the ColoredSurface class which is a pygame Surface."""

from typing import Union
from pygame import Surface, Color, SRCALPHA
from pygame.color import THECOLORS

ColorLike = Union[str, Color, tuple[int, int, int], tuple[int, int, int, int]]

class ColoredSurface(Surface):
    """A ColoredSurface is a Surface with only one color."""

    def __init__(self, color: ColorLike, width: int, height: int):
        super().__init__((width, height), SRCALPHA)

        if isinstance(color, str): # Translate the string into a color
            if color in THECOLORS:
                color = THECOLORS[color]
            elif color.startswith('#'):
                color = Color(color)
            else:
                print(f"'{color}' is not a color, replaced by white.")
                color = Color(255,255,255,255)

        self.fill(color)



"""
The file module contains the File class, which is an abstract base for all file objects,
and the get_file function used in all File class to find the full path of the object
"""
from abc import ABC, abstractmethod
from typing import Any, Literal

import sys
import os
import json

class File(ABC):
    """Represent any type of file that would be loaded for the game: image, sounds, etc."""

    def __init__(self, path: str) -> None:
        self.path = path
        self.full_path = None

    @abstractmethod
    def get(self) -> Any:
        """Get the object in the proper format to be used by the game."""

def get_file(folder: Literal['data', 'musics', 'sounds', 'images', 'videos', 'fonts'], file: str, permanent: bool = False):
    """
    Return the full path of the file.
    
    params:
    ----
    folder: the folder of the file
    file: the name of the file.
    permanent: if False, get the file from the temporary folder of the app. If True, from the place where the file dynamic files are saved
    
    Non-Permanent files are the ones that might be modified during the game and that should be saved.
    Exemple of non-permanent files: saves, ig_queries, logs.
    """
    if folder != 'data':
        folder = 'assets/' + folder
    if hasattr(sys, '_MEIPASS'):
        #pylint: disable=protected-access
        base_path = sys._MEIPASS
        if not permanent:
            config_path = os.path.join(base_path, 'data', 'config.json')
            config_file = open(config_path,'r', encoding='utf-8')
            base_path = json.load(config_file)['path']
            config_file.close()
    else:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, folder, file).replace('\\', '/')



"""The screen is the surface of the window."""

import pygame
from ..settings import Settings
from ..config import Config
from ..file import get_file

class Screen:
    """The screen class is used to represent the screen of the game."""

    def __init__(self, config: Config, settings: Settings) -> None:
        self._width, self._height = config.dimension
        self._fullscreen = False
        self.screen = pygame.display.set_mode((self._width, self._height))
        self._settings = settings

        pygame.display.set_caption(config.game_name)
        pygame.display.set_icon(pygame.image.load(get_file('', 'icon.ico', True)))

    def display_phase(self, phase):
        """Blit the Frame on the screen."""
        self.screen.blit(phase.get_surface(self._width, self._height), (0,0))

    def update(self):
        """Update the screen"""
        if self._fullscreen != self._settings.fullscreen:
            self._fullscreen = self._settings.fullscreen
            if self._fullscreen:
                self.screen = pygame.display.set_mode((self._width, self._height), pygame.FULLSCREEN)
            else:
                self.screen = pygame.display.set_mode((self._width, self._height))

        pygame.display.flip()



"""A MusicFile represent the file of a music."""

import json
from .file import File, get_file

class MusicFile(File):
    """Represent the file of a music in the assets/musics folder."""

    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.full_path = get_file('musics', name)
        with open(get_file('musics', 'loop_times.json'), encoding='utf-8') as f:
            all_loop_times: dict = json.load(f)
            if name in all_loop_times:
                self.loop_time = all_loop_times[name]
            else:
                self.loop_time = None

    def get(self):
        """
        Get the music.
        
        Returns:
        full_path: str, the path to the music
        loop_time: int, the loop time.
        """
        return self.full_path, self.loop_time



"""
The Speeches class is used to manage the speeches of the game by returning SoundFile taking
automatically into account the language, use it with the Soundbox.
"""

from .database import Database
from ..file import SoundFile
from ..settings import Settings
from ..error import PygamingException

class Speeches:
    """
    The class Speeches is used to manage the texts of the game, that might be provided in several languages.
    """

    def __init__(self, database: Database, settings: Settings) -> None:
        self._db = database
        self._settings = settings
        self._last_language = settings.language
        texts_list = self._db.get_texts(self._last_language)
        self._speeches_dict = {pos : txt for pos, txt in texts_list[0]}

    def get_positions(self):
        """Return all the positions (text keys)."""
        return list(self._speeches_dict.keys())

    def get(self, position):
        """Return a path to the speech to be said."""
        if self._settings.language != self._last_language:
            self._last_language = self._settings.language
            speeches_list = self._db.get_texts(self._last_language)
            self._speeches_dict = {pos : spc for pos, spc in speeches_list[0]}

        if position in self._speeches_dict:
            return SoundFile(self._speeches_dict[position]).get()
        raise PygamingException(f"The position {position} does not exist as a speech.")



"""The Jukebox class is used to manage the musics."""

import pygame
from ..file import MusicFile
from ..settings import Settings

_LOOPS = 'loops'
_PLAYLIST = 'playlist'

class Jukebox:
    """The Jukebox is used to manage the musics."""

    def __init__(self, settings: Settings) -> None:

        self._loop_instant = 0
        self._playing = False
        self._settings = settings
        self._loops_or_playlist = _LOOPS
        self._playlist_idx = 0
        self._playlist_playing = []

    def stop(self):
        """Stop the music currently playing."""
        pygame.mixer.music.stop()
        self._playing = False

    def play_loop(self, music_file: MusicFile):
        """Play a music that will loop."""
        path, self._loop_instant = music_file.get()
        self._loops_or_playlist = _LOOPS
        pygame.mixer.music.load(path)
        pygame.mixer.music.play(0)
        self._playing = True
        self._playlist_idx = 0
    
    def read_playlist(self, playlist: list[MusicFile]):
        """Play a playlist"""
        self._loops_or_playlist = _PLAYLIST
        self._playlist_idx = 0
        self._playlist_playing = playlist
        self._playing = True
    
    def add_to_playlist(self, music_file: MusicFile):
        """Add a music to the playlist."""
        self._playlist_playing.append(music_file)

    def update(self):
        """This function should be called at the end of every gameloop to make the music loop or the jukebox play a new music."""
        pygame.mixer.music.set_volume(self._settings.volumes['main']*self._settings.volumes['music'])
        # If we are playing a looping music.
        if self._loops_or_playlist == _LOOPS and not pygame.mixer.music.get_busy() and self._playing and self._loop_instant is not None:
            pygame.mixer.music.play(0, self._loop_instant/1000)

        # If we are reading a playlist
        if self._loops_or_playlist == _PLAYLIST and not pygame.mixer.music.get_busy() and self._playing:
            self._playlist_idx = (self._playlist_idx+1)%len(self._playlist_playing)
            path, _ = self._playlist_playing[self._playlist_idx].get()
            pygame.mixer.music.load(path)
            pygame.mixer.music.play(0)



"""The data submodule of file is an abstract class used to represent the different types of data stored by the game."""
from abc import ABC, abstractmethod
from .file import File, get_file

class DataFile(File, ABC):
    """
    Abstract class for files of any type of data.
    Use this class to create file that are specific for the
    different type of data you store.
    e.g: player stored as .json, stats stored as .csv,
    places, objects, environements stored with custom extensions...
    """

    def __init__(self, path: str, dynamic: bool = False) -> None:
        File.__init__(self, path)
        ABC.__init__(self)
        self.full_path = get_file('data', path, dynamic)

    @abstractmethod
    def get(self):
        """Return the object transformed with the file."""



"""
The function init_cwd is used to create the working environment first.
It should be only used by using the command line defined in the `setup.py` file: pygaming init
The function make some basic folders and files to assure the localization of the files for the building
and to guide the programmer.
"""

import os
import shutil
import json
import locale

def init_cwd():
    """Create some files and folder in the current working directory."""

    cwd = os.getcwd()
    this_dir = os.path.dirname(__name__)

    if not os.path.exists(os.path.join(cwd, 'assets')):

        shutil.copytree(
            os.path.join(this_dir, 'pygaming', 'commands/templates/assets'),
            os.path.join(cwd, 'assets')
        )

        # Print the success output ot guide the user
        with open(os.path.join(this_dir, 'pygaming', 'commands/init_texts/assert_success.txt'), 'r', encoding='utf-8') as f:
            text = ''.join(f.readlines())
        print("\033[33m" + text + "\033[0m")

    if not os.path.exists(os.path.join(cwd, 'data')):

        language = locale.getlocale()[0]

        shutil.copytree(
            os.path.join(this_dir, 'pygaming', 'commands/templates/data'),
            os.path.join(cwd, 'data')
        )

        with open(os.path.join(cwd, 'data/settings.json'), 'r', encoding='utf-8') as f:
            settings = json.load(f)
            settings['current_languge'] = language
        with open(os.path.join(cwd, 'data/settings.json'), 'w', encoding='utf-8') as f:
            json.dump(settings, f)

        with open(os.path.join(cwd, 'data/config.json'), 'r', encoding='utf-8') as f:
            settings = json.load(f)
            settings['current_languge'] = language
        with open(os.path.join(cwd, 'data/config.json'), 'w', encoding='utf-8') as f:
            json.dump(settings, f)


        with open(os.path.join(this_dir, 'pygaming', 'commands/init_texts/data_success.txt'), 'r', encoding='utf-8') as f:
            text = ''.join(f.readlines())
        print("\033[35m" + text + "\033[0m")


    if not os.path.exists(os.path.join(cwd, 'src')):

        shutil.copytree(
            os.path.join(this_dir, 'pygaming', 'commands/templates/src'),
            os.path.join(cwd, 'src')
        )

        with open(os.path.join(this_dir, 'pygaming', 'commands/init_texts/src_success.txt'), 'r', encoding='utf-8') as f:
            text = ''.join(f.readlines())

        print("\033[32m" + text + "\033[0m")

    with open(os.path.join(this_dir, 'pygaming', 'commands/init_texts/init_success.txt'), 'r', encoding='utf-8') as f:
        text = ''.join(f.readlines())

    # Reinitialize the color text and print the end of the init text
    print(text)



"""Font files represent the file of a font."""
import pygame

from .file import File, get_file
from ..font import Font

class FontFile(File):
    """Represent the file of a font stored in the assets/font folder."""

    def __init__(self, name: str) -> None:
        super().__init__(name)
        if name.endswith('.ttf') or name.endswith('.TTF'):
            self.full_path = get_file(folder='fonts', file=name)
        else:
            self.full_path = get_file(folder='fonts', file=name + '.ttf')

    #pylint: disable=arguments-differ
    def get(
            self,
            size: int,
            color: pygame.Color,
            settings,
            italic: bool = False,
            bold: bool = False,
            underline: bool = False
        ) -> pygame.font.Font:
        """
        Get the font
        
        Params:
        ----
        size: int, the size of the font.
        color: pygame.Color: the color of the Font
        settings: Settings, the self.settings of the Game
        italic, bold, underline: bools, flags for the font.
        """
        return Font(self.full_path, size, color, settings, bold, italic, underline)

default_font = FontFile("")
default_font.full_path = None



from setuptools import setup, find_packages

setup(
    name='pygaming',
    author="Tanguy Dugas du Villard",
    author_mail="tanguy.dugas01@gmail.com",
    version='0.2.0',
    description="Pygaming is a python library based on pygame used to create game more easily by providing several tools.",
    packages=find_packages(),
    install_requires=[
        'pyinstaller',
        'pygame',
        'numpy',
        'pillow'
    ],
    entry_points={
        'console_scripts': [
            'pygaming=pygaming.commands.cli:cli'
        ],
    },
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Tanguy-ddv/pygaming/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True,
    package_data={
        '': ['commands/templates/*'],
    },
)



"""The settings class is used to interact with the settings file."""
import json
from typing import Any
from .file import get_file
from .error import PygamingException

class Settings:
    """The settings class is used to interact with the settings file."""

    def __init__(self) -> None:
        self._path = get_file('data', 'settings.json', False)
        file = open(self._path, 'r', encoding='utf-8')
        self._data = json.load(file)
        file.close()

    def get(self, attribute: str):
        """Get the value of the settings attribute"""
        if attribute in self._data:
            return self._data[attribute]
        return None

    @property
    def language(self) -> str:
        """Return the current language setting."""
        return self._data['current_language']

    @property
    def antialias(self) -> bool:
        """Return the antialias setting."""
        if 'antialias' in self._data:
            return self._data['antialias']
        return True

    @property
    def fullscreen(self) -> bool:
        """Return the fullscreen setting."""
        if "fullscreen" in self._data:
            return self._data['fullscreen']
        return False

    @property
    def controls(self) -> dict[str, str]:
        """Return the controls."""
        return self._data['controls']

    @property
    def volumes(self) -> dict[str, Any]:
        """Return the volumes."""
        return self._data['volumes']

    def save(self) -> None:
        """Save the current settings."""
        file = open(self._path, 'w', encoding='utf-8')
        json.dump(self._data, file)
        file.close()

    def set_volumes(self, volumes: dict[str, Any]):
        """
        Set the volumes to new values.
        The new volumes must be a dict with keys: 'main', 'sounds' and 'music',
        with 'main' and 'music' mapping to a number between 0 and 1
        and 'sounds' mapping to a dict of category : number between 0 and 1.
        The categories of the new volumes and the previous volume must be the exact same.
        """
        if not ('main' in volumes and 'sounds' in volumes and 'music' in volumes):
            raise PygamingException("'main', 'sounds' or 'music' is not present in the new volumes dict.")
        for key in self._data['volumes']['sounds']:
            if key not in volumes['sounds']:
                raise PygamingException(f"The category {key} is not present in the new sounds volumes.")
        for key in volumes['sounds']:
            if key not in self._data['volumes']['sounds']:
                raise PygamingException(f"The category {key} is not defined in the settings files.")
        self._data['volumes'] = volumes
        self.save()

    def set_language(self, language: str):
        """Set the new language."""
        self._data['current_language'] = language
        self.save()

    def set_controls(self, controls: dict[str, str]):
        """Set the new keymap."""
        for key in self.controls.values():
            if key not in controls.values():
                raise PygamingException(f"the action {key} have not been mapped.")
        for key in controls.values():
            if key not in self.controls.values():
                raise PygamingException(f"the action {key} does not exists.")
        self._data['controls'] = controls
        self.save()

    def set_full_screen(self, fullscreen : bool):
        """Set the full screen."""
        self._data['fullscreen'] = fullscreen
        self.save()

    def set_attribute(self, attribute: str, value: Any):
        """Set the new value for a given attribute."""
        if attribute not in self._data:
            raise PygamingException(f"{attribute} is not an attribute of the settings.")
        if attribute in ['volumes', 'current_language', 'full_screen', 'controls']:
            raise PygamingException(f"Please set {attribute} with it dedecated setter.")
        self._data[attribute] = value
        self.save()



"""The logger is used to log in-game actions to replay matches after."""

import time
import os
import json
from datetime import datetime
from .file import get_file
from .config import Config

class Logger:
    """
    A logger is used to store the log of a game.
    It might be used to compute statitics, replay actions, ...
    Logs are stored as "data/logs/'timestamp'.log"
    """
    def __init__(self, config: Config, debug: bool = False) -> None:

        self.timestamp = datetime.now().strftime("%Y-%m-%d %H-%M-%S")
        self.time_since_last_flush = 0
        self.config = config
        log_dir = get_file('data', 'logs', permanent=False)
        self.debug = debug
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)
        self._file = open(get_file('data', f'logs/{self.timestamp}.log', True), 'a', encoding='utf-8') # pylint: disable=consider-using-with
        # We use this to get the log file always open to save performance with openning and closing.

    def write(self, data: dict, is_it_debugging: bool = False):
        """Write a new line in the log."""
        if self.debug or not is_it_debugging:
            data['timestamp'] = time.time_ns()
            json.dump(data, self._file)

    def update(self, loop_duration: int):
        """Update the logger at every iteration to flush regularly."""
        self.time_since_last_flush += loop_duration
        if self.time_since_last_flush > self.config.get("flush_frequency"):
            self._file.flush()

    def new_log(self):
        """Start a new log."""
        self._file.close()
        self.timestamp = datetime.now().strftime("%Y-%m-%d, %H:%M:%S")
        self._file = open(get_file('data', f'logs/{self.timestamp}.log', True), 'a', encoding='utf-8')

    def __del__(self):
        """At the end of the game, close the file."""
        self._file.close()



"""AnimatedSurface is a class use to represent the a Surface that might be animated."""
from typing import Union, List, Iterable
import pygame
from ..file import ImageFile, GIFFile
from ..error import PygamingException
from .colored_surface import ColoredSurface

SurfaceLike = Union[pygame.Surface, ColoredSurface, ImageFile]
SurfacesLike = Union[List[SurfaceLike], GIFFile]

class AnimatedSurface:
    """An AnimatedSurface is a Surface that can be animated."""

    def __init__(
        self,
        surfaces: SurfacesLike,
        image_duration: int | list[int] = 100, # [ms]
        image_introduction: int = 0
    ) -> None:
        """
        Create the surfaces

        params:
        ----
        width: int, the width of the object.
        height: int, the hieght of the object.
        surfaces: SurfacesLike, The surfaces of the objects.
        if only one element is given, it is treated as a list of length 1
        If it is a (list of) color or a str, create a list of surface of this color with the shape (width, height)
        If it is a (list of) surface, resize the surface with (width, height)
        Can be a list of colors and surfaces, str
        image_duration: If several surfaces are given, as a list of str, color, ImageFile or Surface,
        the frame duration is the amount of time each frame is displayed before. If it is a list, it must be the same length than surfaces.
        image_introduction: int, default 0. If an integer is given (< length of surfaces), the loop does not go back to the first image but to this one.
        ex: In a platformer, the 5 first frames are the character standing in the right direction, then he walks. For this, we use a image_introduction=5
        """
        self._index = 0
        self._image_introduction = image_introduction
        self._introduction_done = False
        self._time_since_last_change = 0

        if isinstance(surfaces, GIFFile):
            surfaces, image_duration = surfaces.get()
        if not isinstance(surfaces, Iterable):
            raise PygamingException("The surface must be an iterable or a GIFFile")

        self._surfaces: list[pygame.Surface] = []
        for bg in surfaces:
            if isinstance(bg, ImageFile):
                bg = bg.get()
            self._surfaces.append(bg)
        self._n_bg = len(surfaces)

        if not isinstance(image_duration, Iterable):
            image_duration = [image_duration]*self._n_bg
        elif len(image_duration) != self._n_bg:
            raise PygamingException(
                f"The length of the frame duration list ({len(image_duration)}) does not match the len of the backroung list ({self._n_bg}))"
            )
        self._image_durations = image_duration
        if self._image_introduction > self._n_bg:
            raise PygamingException(
                f"The image introduction parameters must be between 0 and {self._n_bg}, but got {self._image_introduction}"
            )

    def update_animation(self, loop_duration: float):
        """Update the background"""
        if self._n_bg > 1:
            self._time_since_last_change += loop_duration
            if self._time_since_last_change >= self._image_durations[self._index]:
                self._time_since_last_change = 0
                if not self._introduction_done:
                    self._index = (self._index+1)%self._n_bg
                    if self._index > self._image_introduction:
                        self._introduction_done = True
                else:
                    self._index = (self._index+1 - self._image_introduction)%(self._n_bg - self._image_introduction) + self._image_introduction
                    print(self._index)

    @property
    def height(self):
        """Return the height of the surface, which is the height of the first surface."""
        return self._surfaces[0].get_height()

    @property
    def width(self):
        """Return the width of the surface, which is the width of the first surface."""
        return self._surfaces[0].get_width()

    def reset(self):
        """Reset the counts of the animations."""
        self._index = 0
        self._introduction_done = False
        self._time_since_last_change = 0

    def get(self):
        """
        Return the background.
        """
        self._index = self._index%self._n_bg
        return self._surfaces[self._index].copy()

def make_rounded_rectangle(color: pygame.Color | str, width: int, height: int):
    """Make a rectange with half circles at the start and end."""
    if isinstance(color, str):
        if color in pygame.color.THECOLORS:
            color = pygame.color.THECOLORS[color]
        else:
            color = pygame.Color(0,0,0,255)

    surface = pygame.Surface((width, height), pygame.SRCALPHA)
    rect = pygame.Rect(height//2, 0, width - height, height)
    pygame.draw.rect(surface, color, rect)
    pygame.draw.circle(surface, color, (height//2, height//2), height//2)
    pygame.draw.circle(surface, color, (width - height//2, height//2), height//2)
    return surface
